import logging
import os
import shutil
from copy import deepcopy
from datetime import datetime
from textwrap import indent
from typing import Generator, Iterable, List, NamedTuple, Optional, Sequence, Set, Tuple

from colorama import Fore, Style  # type: ignore

from .actions.action import Action
from pathlib import Path
from .config import Rule
from .filters.filter import Filter
from .utils import DotDict, splitglob

logger = logging.getLogger(__name__)
SYSTEM_FILES = ("thumbs.db", "desktop.ini", ".DS_Store")

Job = NamedTuple(
    "Job",
    [
        ("folderstr", str),
        ("basedir", Path),
        ("path", Path),
        ("filters", Sequence[Filter]),
        ("actions", Sequence[Action]),
    ],
)
Job.__doc__ = """
    :param str folderstr: the original folder definition specified in the config
    :param Path basedir:  the job's base folder
    :param Path path:     the path of the file to handle
    :param list filters:  the filters that apply to the path
    :param list actions:  the actions which should be executed
"""


class OutputHelper:
    """
    class to track the current folder / file and print only changes.
    This is needed because we only want to output the current folder and file if the
    filter or action prints something.
    """

    def __init__(self) -> None:
        self.not_found = set()  # type: Set[str]
        self.curr_folder = None  # type: Optional[Path]
        self.curr_path = None  # type: Optional[Path]
        self.prev_folder = None  # type: Optional[Path]
        self.prev_path = None  # type: Optional[Path]

    def set_location(self, folder: Path, path: Path) -> None:
        self.curr_folder = folder
        self.curr_path = path

    def pre_print(self) -> None:
        """
        pre-print hook that is called everytime the moment before a filter or action is
        about to print something to the cli
        """
        if self.curr_folder != self.prev_folder:
            if self.prev_folder is not None:
                print()  # ensure newline between folders
            print("Folder %s%s:" % (Style.BRIGHT, self.curr_folder))
            self.prev_folder = self.curr_folder

        if self.curr_path != self.prev_path:
            print(indent("File %s%s:" % (Style.BRIGHT, self.curr_path), " " * 2))
            self.prev_path = self.curr_path

    def print_path_not_found(self, folderstr: str) -> None:
        if folderstr not in self.not_found:
            self.not_found.add(folderstr)
            msg = "Path not found: {}".format(folderstr)
            print(Fore.YELLOW + Style.BRIGHT + msg)
            logger.warning(msg)


output_helper = OutputHelper()


def execute_rules(rules: Iterable[Rule], simulate: bool) -> None:
    cols, _ = shutil.get_terminal_size(fallback=(79, 20))
    simulation_msg = Fore.GREEN + Style.BRIGHT + " SIMULATION ".center(cols, "~")

    jobs = create_jobs(rules=rules)

    if simulate:
        print(simulation_msg)

    failed, succeded = run_jobs(jobs=jobs, simulate=simulate)
    if succeded == failed == 0:
        msg = "Nothing to do."
        logger.info(msg)
        print(msg)

    if simulate:
        print(simulation_msg)


def create_jobs(rules: Iterable[Rule]) -> Generator[Job, None, None]:
    """ creates `Job` data structures for every path handled in each rule """
    for rule in rules:
        for folderstr, basedir, path in all_files_for_rule(rule):
            yield Job(
                folderstr=folderstr,
                basedir=basedir,
                path=path,
                filters=rule.filters,
                actions=rule.actions,
            )


def all_files_for_rule(rule: Rule) -> Generator[Tuple[str, Path, Path], None, None]:
    files = dict()
    for folderstr in rule.folders:
        folderstr = folderstr.strip()

        # check whether the file / folder is prefixed with `!` to be excluded
        exclude_flag = folderstr.startswith("!")

        # assemble glob expression
        basedir, globstr = splitglob(folderstr.lstrip("!").strip())
        if basedir.is_dir():
            if not globstr:
                globstr = "**/*" if rule.subfolders else "*"
        elif basedir.is_file():
            # this allows specifying single files
            globstr = basedir.name
            basedir = basedir.parent
        else:
            output_helper.print_path_not_found(str(basedir))
            continue

        # iterate files in basedir and add to / remove from result dict
        for path in basedir.glob(globstr):
            if path.is_file() and (rule.system_files or path.name not in SYSTEM_FILES):
                if not exclude_flag:
                    files[path] = (folderstr, basedir)
                elif path in files:
                    del files[path]

    for path, (folderstr, basedir) in files.items():
        yield (folderstr, basedir, path)


def run_jobs(jobs: Iterable[Job], simulate: bool) -> List[int]:
    """ :returns: The number of successfully handled files """
    count = [0, 0]
    Action.pre_print_hook = output_helper.pre_print
    Filter.pre_print_hook = output_helper.pre_print

    for job in sorted(jobs, key=lambda x: (x.folderstr, x.basedir, x.path)):
        args = DotDict(
            path=job.path,
            basedir=job.basedir,
            simulate=simulate,
            relative_path=job.path.relative_to(job.basedir),
            env=os.environ,
            now=datetime.now(),
        )

        output_helper.set_location(job.basedir, args.relative_path)
        match = filter_pipeline(filters=job.filters, args=args)
        if match:
            success = action_pipeline(actions=job.actions, args=args)
            count[success] += 1
    return count


def filter_pipeline(filters: Iterable[Filter], args: DotDict) -> bool:
    """
    run the filter pipeline.
    Returns True on a match, False otherwise and updates `args` in the process.
    """
    for filter_ in filters:
        try:
            result = filter_.pipeline(deepcopy(args))
            if isinstance(result, dict):
                args.update(result)
            elif not result:
                # filters might return a simple True / False.
                # Exit early if a filter does not match.
                return False
        except Exception as e:  # pylint: disable=broad-except
            logger.exception(e)
            filter_.print(Fore.RED + Style.BRIGHT + "ERROR! %s" % e)
            return False
    return True


def action_pipeline(actions: Iterable[Action], args: DotDict) -> bool:
    for action in actions:
        try:
            updates = action.pipeline(deepcopy(args))
            # jobs may return a dict with updates that should be merged into args
            if updates is not None:
                args.update(updates)
        except Exception as e:  # pylint: disable=broad-except
            logger.exception(e)
            action.print(Fore.RED + Style.BRIGHT + "ERROR! %s" % e)
            return False
    return True



import mimetypes

from pathlib import Path
from organize.utils import DotDict, flatten

from .filter import Filter


class MimeType(Filter):

    """
    Filter by MIME type associated with the file extension.

    Supports a single string or list of MIME type strings as argument.
    The types don't need to be fully specified, for example "audio" matches everything
    from "audio/midi" to "audio/quicktime".

    You can see a list of known MIME types on your system by running this oneliner:

    .. code-block:: yaml

      python3 -c "import mimetypes as m; print('\\n'.join(sorted(set(m.common_types.values()) | set(m.types_map.values()))))"


    Examples:
        - Show MIME types:

          .. code-block:: yaml
            :caption: config.yaml

            rules:
              - folders: '~/Downloads'
                filters:
                  - mimetype
                actions:
                  - echo: '{mimetype}'

        - Filter by "image" mimetype:

          .. code-block:: yaml
            :caption: config.yaml

            rules:
              - folders: '~/Downloads'
                filters:
                  - mimetype: image
                actions:
                  - echo: This file is an image: {mimetype}

        - Filter by specific MIME type:

          .. code-block:: yaml
            :caption: config.yaml

            rules:
              - folders: '~/Desktop'
                filters:
                  - mimetype: application/pdf
                actions:
                  - echo: 'Found a PDF file'

        - Filter by multiple specific MIME types:

          .. code-block:: yaml
            :caption: config.yaml

            rules:
              - folders: '~/Music'
                filters:
                  - mimetype:
                    - application/pdf
                    - audio/midi
                actions:
                  - echo: 'Found Midi or PDF.'
    """

    def __init__(self, *mimetypes):
        self.mimetypes = list(map(str.lower, flatten(list(mimetypes))))

    @staticmethod
    def mimetype(path):
        type_, _ = mimetypes.guess_type(path, strict=False)
        return type_

    def matches(self, path: Path):
        mimetype = self.mimetype(path)
        if mimetype is None:
            return False
        if not self.mimetypes:
            return True
        return any(mimetype.startswith(x) for x in self.mimetypes)

    def pipeline(self, args: DotDict):
        if self.matches(args.path):
            result = self.mimetype(args.path)
            return {"mimetype": result}
        return None

    def __str__(self):
        return "MimeType(%s)" % ", ".join(self.mimetypes)



import logging
import sys
from pathlib import Path
from typing import Mapping

import simplematch as sm # type: ignore

from .action import Action

logger = logging.getLogger(__name__)


class MacOSTags(Action):

    """
    Add macOS tags.

    Example:
        - Add a single tag:

          .. code-block:: yaml
            :caption: config.yaml

            rules:
              - folders: '~/Documents/Invoices'
              - filters:
                  - filename:
                      startswith: "Invoice"
                  - extension: pdf
              - actions:
                  - macos_tags: Invoice

        - Adding multiple tags ("Invoice" and "Important"):

          .. code-block:: yaml
            :caption: config.yaml

            rules:
              - folders: '~/Documents/Invoices'
              - filters:
                  - filename:
                      startswith: "Invoice"
                  - extension: pdf
              - actions:
                  - macos_tags:
                    - Important
                    - Invoice

        - Specify tag colors. Available colors are `none`, `gray`, `green`, `purple`, `blue`, `yellow`, `red`, `orange`.

          .. code-block:: yaml
            :caption: config.yaml

            rules:
              - folders: '~/Documents/Invoices'
              - filters:
                  - filename:
                      startswith: "Invoice"
                  - extension: pdf
              - actions:
                  - macos_tags:
                    - Important (green)
                    - Invoice (purple)

        - Add a templated tag with color:

          .. code-block:: yaml
            :caption: config.yaml

            rules:
              - folders: '~/Documents/Invoices'
              - filters:
                  - created
              - actions:
                  - macos_tags:
                    - Year-{created.year} (red)
    """

    def __init__(self, *tags):
        self.tags = tags

    def pipeline(self, args: Mapping):
        path = args["path"]  # type: Path
        simulate = args["simulate"]  # type: bool

        if sys.platform != "darwin":
            self.print("The macos_tags action is only available on macOS")
            return

        import macos_tags  # type: ignore

        COLORS = [c.name.lower() for c in macos_tags.Color]

        for template in self.tags:
            tag = self.fill_template_tags(template, args)
            name, color = self._parse_tag(tag)

            if color not in COLORS:
                raise ValueError(
                    "color %s is unknown. (Available: %s)" % (color, " / ".join(COLORS))
                )

            self.print('Adding tag: "%s" (color: %s)' % (name, color))
            if not simulate:
                _tag = macos_tags.Tag(
                    name=name,
                    color=macos_tags.Color[color.upper()],
                )
                macos_tags.add(_tag, file=str(path))

    def _parse_tag(self, s):
        """ parse a tag definition and return a tuple (name, color) """
        result = sm.match("{name} ({color})", s)
        if not result:
            return s, "none"
        return result["name"], result["color"].lower()

    def __str__(self) -> str:
        return "MacOSTags(tags=%s)" % self.tags



"""
Duplicate detection filter.

Based on this stackoverflow answer:
    https://stackoverflow.com/a/36113168/300783

Which was updated for python3 in:
    https://gist.github.com/tfeldmann/fc875e6630d11f2256e746f67a09c1ae

The script on stackoverflow has a bug which could lead to false positives. This is fixed
here by using a tuple (file_size, hash) as key in the comparison dictionaries.
"""
import hashlib
import os
from collections import defaultdict
from typing import DefaultDict as DDict
from typing import Dict, List, Set, Tuple, Union

from organize.utils import fullpath

from .filter import Filter


def chunk_reader(fobj, chunk_size=1024):
    """ Generator that reads a file in chunks of bytes """
    while True:
        chunk = fobj.read(chunk_size)
        if not chunk:
            return
        yield chunk


def get_hash(filename, first_chunk_only=False, hash_algo=hashlib.sha1):
    hashobj = hash_algo()
    with open(filename, "rb") as f:
        if first_chunk_only:
            hashobj.update(f.read(1024))
        else:
            for chunk in chunk_reader(f):
                hashobj.update(chunk)
    return hashobj.digest()


class Duplicate(Filter):

    """
    Finds duplicate files.

    This filter compares files byte by byte and finds identical files with potentially
    different filenames.

    :returns:
        - ``{duplicate}`` -- path to the duplicate source

    Examples:
        - Show all duplicate files in your desktop and download folder (and their
          subfolders).

          .. code-block:: yaml
            :caption: config.yaml

            rules:
            - folders:
                - ~/Desktop
                - ~/Downloads
                subfolders: true
                filters:
                - duplicate
                actions:
                - echo: "{path} is a duplicate of {duplicate}"
    """

    def __init__(self) -> None:
        self.files_for_size = defaultdict(list)  # type: DDict[int, List[str]]

        # to prevent false positives the keys must be tuples of (file_size, hash).
        self.files_for_small_hash = defaultdict(
            list
        )  # type: DDict[Tuple[int, bytes], List[str]]
        self.file_for_full_hash = dict()  # type: Dict[Tuple[int, bytes], str]

        # we keep track of which files we already computed the hashes for so we only do
        # that once.
        self.small_hash_known = set()  # type: Set[str]
        self.full_hash_known = set()  # type: Set[str]

    def matches(self, path: str) -> Union[bool, Dict[str, str]]:
        # the exact same path has already been handled. This might happen if path is a
        # symlink which resolves to file that is already known. We skip these.
        if path in self.small_hash_known:
            return False

        # check for files with equal size
        file_size = os.path.getsize(path)  # type: int
        same_size = self.files_for_size[file_size]
        candidates_fsize = same_size[:]
        same_size.append(path)
        if not candidates_fsize:
            # the file is unique in size and cannot be a duplicate
            return False

        # for all other files with the same file size, get their hash of the first 1024
        # bytes
        for c in candidates_fsize:
            if c not in self.small_hash_known:
                try:
                    c_small_hash = get_hash(c, first_chunk_only=True)
                    self.files_for_small_hash[(file_size, c_small_hash)].append(c)
                    self.small_hash_known.add(c)
                except OSError:
                    pass

        # check small hash collisions with the current file
        small_hash = get_hash(path, first_chunk_only=True)
        same_small_hash = self.files_for_small_hash[(file_size, small_hash)]
        candidates_shash = same_small_hash[:]
        same_small_hash.append(path)
        self.small_hash_known.add(path)
        if not candidates_shash:
            # the file has a unique small hash and cannot be a duplicate
            return False

        # For all other files with the same file size and small hash get the full hash
        for c in candidates_shash:
            if c not in self.full_hash_known:
                try:
                    c_full_hash = get_hash(c, first_chunk_only=False)
                    self.file_for_full_hash[(file_size, c_full_hash)] = c
                    self.full_hash_known.add(c)
                except OSError:
                    pass

        # check full hash collisions with the current file
        full_hash = get_hash(path, first_chunk_only=False)
        duplicate = self.file_for_full_hash.get((file_size, full_hash))
        if duplicate:
            return {"duplicate": duplicate}
        self.file_for_full_hash[(file_size, full_hash)] = path
        return False

    def pipeline(self, args):
        return self.matches(str(fullpath(args["path"])))

    def __str__(self) -> str:
        return "Duplicate()"



from .copy import Copy
from .delete import Delete
from .echo import Echo
from .macos_tags import MacOSTags
from .move import Move
from .python import Python
from .rename import Rename
from .shell import Shell
from .trash import Trash
